#Get the categories for the catNeurons.
#Use those to categorise a run. 
import pickle
import sys
import numpy as np

EPOCH_LENGTH = 400
defaults = 0

def getNeuronCats(fileName):
        fileHandle = open(fileName,'rb')
        neuronCats = pickle.load(fileHandle)
        fileHandle.close()
        return (neuronCats)


def readSpikes (fileName):
    spikes = []
    fileHandle = open(fileName,'r')
    line = fileHandle.readline()
    while not (line == ""):
        arguments = line.split()
        neuronNumber = int(arguments[0])
        time = float(arguments[1])
        spike = (neuronNumber,time)
        spikes = spikes + [spike]
        line = fileHandle.readline()
    fileHandle.close()

    return spikes

def getSpikesPerNeuronPerEpoch(spikes,startTime,endTime):
    neuronSpikeCounts = np.zeros(100)
    for spike in spikes:
        neuron = spike[0]
        time = spike[1]
        if (time >= startTime) and (time < endTime):
                neuronSpikeCounts[neuron] +=1
    return (neuronSpikeCounts)

def oddResolve(numberZeros,numberOnes,numberTwos):
    if (numberZeros > numberOnes):
            answerCategory = 0
    elif (numberZeros < numberOnes):
            answerCategory = 1
    else:
            answerCategory = 2
    return (answerCategory)

def defaultResolve(numberZeros,numberOnes,numberTwos):
    if (numberZeros == 0) and (numberOnes == 0) and (numberTwos == 0):
            answerCategory = 0
    elif (numberZeros > numberOnes):
            answerCategory = 0
    elif (numberZeros < numberOnes):
            answerCategory = 1
    else:
            answerCategory = 0
    return (answerCategory)


def fairResolve(numberZeros,numberOnes,numberTwos):
    global defaults
    if (numberZeros > numberOnes) and (numberZeros>numberTwos):
            answerCategory = 0
    elif (numberOnes > numberZeros) and (numberOnes>numberTwos):
            answerCategory = 1
    elif (numberTwos > numberZeros) and (numberTwos>numberOnes):
            answerCategory = 2
    #elif (numberTwos == numberOnes) and (numberTwos>numberZeros):
    #        answerCategory = 2
    #elif (numberTwos == numberZeros) and (numberTwos>numberOnes):
    #        answerCategory = 0
    else: # default
            print("default")
            defaults += 1
            answerCategory = 0
    return (answerCategory)


def getCategory(neuronCats,neuronSpikeCounts):
    answerCategory = -1
    numberZeros = 0.0
    numberOnes = 0.0
    numberTwos = 0.0
    #get the contribution per category per testSpike
    for neuronNumber in range (0,100):
        numberSpikes = neuronSpikeCounts[neuronNumber]
        #neurons are waited negatively by the number of training spikes
        divisor = numberSpikes*neuronCats[neuronNumber][3]
        if ((divisor > 0) and (numberSpikes > 0)):
            cat0 = numberSpikes*neuronCats[neuronNumber][0]/divisor
            numberZeros += cat0
            cat1 = numberSpikes*neuronCats[neuronNumber][1]/divisor
            numberOnes += cat1
            cat2 = numberSpikes*neuronCats[neuronNumber][2]/divisor
            numberTwos += cat2
    
    #print (str(numberZeros) + " " +str(numberOnes) + " " +str(numberTwos))
    zeroNormalizer = neuronCats[100][3]/neuronCats[100][0]
    oneNormalizer = neuronCats[100][3]/neuronCats[100][1]
    twoNormalizer = neuronCats[100][3]/neuronCats[100][2]
    numberZeros *= zeroNormalizer
    numberOnes *= oneNormalizer
    numberTwos *= twoNormalizer
    print (str(numberZeros) + " " +str(numberOnes) + " " +str(numberTwos))

    #answerCategory = oddResolve(numberZeros,numberOnes,numberTwos)
    #answerCategory = defaultResolve(numberZeros,numberOnes,numberTwos)
    answerCategory = fairResolve(numberZeros,numberOnes,numberTwos)
    
    return (answerCategory)
        
        
def categorise (neuronCats,allSpikes):
    numberCorrect = 0
    numberZeros = 0
    numberOnes = 0
    numberTwos = 0
    for epoch in range (0,75):
        startTime = epoch*EPOCH_LENGTH
        endTime = startTime + EPOCH_LENGTH
        neuronSpikeCounts = getSpikesPerNeuronPerEpoch(allSpikes,startTime,endTime)
        #print (neuronSpikeCounts)

        category = getCategory(neuronCats,neuronSpikeCounts)
        correctAnswer = epoch%3
        if (category == correctAnswer):
                numberCorrect += 1
        if (category == 0):
                numberZeros += 1
        elif (category == 1):
                numberOnes += 1
        elif (category == 2):
                numberTwos += 1

        #print (str(category) + " " + str(correctAnswer) + " =" + str(numberCorrect))
        print (str(numberZeros) + "/" + str(numberOnes) + "/" + str(numberTwos))

    return numberCorrect

#--main
args = sys.argv
numberArgs = args.__len__()
if (numberArgs == 2):
    fileName = args[1]
elif (numberArgs == 3):
    fileName = args[1]
    spikeFileName = args[2]
else:
    fileName = "spikesPerCat.pkl"
    spikeFileName = "cat.sp"

neuronCats = getNeuronCats(fileName)
print (neuronCats)
allSpikes = readSpikes(spikeFileName)
result = categorise (neuronCats,allSpikes)
print ("defaults " + str(defaults))

print (result)
