#Get the category spike ratios for the category neurons.
#Example call
#python categoriseMNIST.py  spikesPerMNISTCat.pkl fiveACat.sp inputData/five-six-two.csv
#Use those to categorise a run. 
import pickle
import sys
import numpy as np

EPOCH_LENGTH = 400
#NUMBER_EPOCHS = 281
NUMBER_EPOCHS = 562
#NUMBER_CAT_NEURONS = 100
NUMBER_CAT_NEURONS = 200

defaults = 0

def processLine(lineString):
    dataLine = lineString.split(',')
    #print (dataLine)
    #print (len(dataLine))
    label = int (dataLine[len(dataLine)-1])
    #print(label)
    features = []
    for featureNumber in range (0,len(dataLine)-1):
        nextFeature = int(dataLine[featureNumber])
        features = features + [nextFeature]
        
    label = [label]

    #print(features)
    #print(label)
    
    return (features,label)

#Read in the mnist csv file returning a data matrix and label vector.
def readMNISTDataFile(fileName):
    fileHandle = open(fileName,'r')
    features = []
    labels = []
    for line  in fileHandle:
        feature,label = processLine(line)
        features = features + [feature]
        labels = labels + label
    fileHandle.close()

    #print(labels)
    #print(len(labels))
    return(features,labels)

def getNeuronCatMatrix(fileName):
        fileHandle = open(fileName,'rb')
        neuronCats = pickle.load(fileHandle)
        fileHandle.close()
        return (neuronCats)

#add the sum column and sum row to the matrix
def addSums(neuronCatMatrix):
        newMatrix = np.zeros((NUMBER_CAT_NEURONS+1,11))

        #copy the old matrix to the new larger one
        for neuron in range (0,NUMBER_CAT_NEURONS):
             for cat in range (0,10):
                    newMatrix[neuron][cat] = neuronCatMatrix[neuron][cat]

        #add in the total column
        for neuron in range (0,NUMBER_CAT_NEURONS):
             totalNeuronSpikes = 0
             for cat in range (0,10):
                     totalNeuronSpikes += newMatrix[neuron][cat]
             newMatrix[neuron][10] = totalNeuronSpikes

        #add in the total row
        totalMatrixSpikes = 0
        for cat in range (0,10):
             totalNeuronSpikes = 0
             for neuron in range (0,NUMBER_CAT_NEURONS):
                     totalNeuronSpikes += newMatrix[neuron][cat]
             newMatrix[NUMBER_CAT_NEURONS][cat] = totalNeuronSpikes
             totalMatrixSpikes += totalNeuronSpikes
        
        newMatrix[NUMBER_CAT_NEURONS][10] = totalMatrixSpikes

        return(newMatrix)


def readSpikes (fileName):
    spikes = []
    fileHandle = open(fileName,'r')
    line = fileHandle.readline()
    while not (line == ""):
        arguments = line.split()
        neuronNumber = int(arguments[0])
        time = float(arguments[1])
        spike = (neuronNumber,time)
        spikes = spikes + [spike]
        line = fileHandle.readline()
    fileHandle.close()

    return spikes

def getSpikesPerNeuronPerEpoch(spikes,startTime,endTime):
    neuronSpikeCounts = np.zeros(NUMBER_CAT_NEURONS)
    for spike in spikes:
        neuron = spike[0]
        time = spike[1]
        if (time >= startTime) and (time < endTime):
                neuronSpikeCounts[neuron] +=1
    return (neuronSpikeCounts)

def resolve(catWeights):
    global defaults
    largest = -1.0
    answerCategory = -1
    for cat in range (0,10):
            if(catWeights[cat] > largest):
                    answerCategory = cat
                    largest = catWeights[cat]
    if (largest == 0.0):
            defaults += 1
            answerCategory = 1

    return (answerCategory)


def getCategory(neuronCatMatrix,neuronSpikeCounts):
    #print(neuronCatMatrix)
    #print(neuronSpikeCounts)
    answerCategory = -1
    #get the contribution per category per testSpike
    catWeights = np.zeros(10)
    for neuronNumber in range (0,NUMBER_CAT_NEURONS):
        numberSpikes = neuronSpikeCounts[neuronNumber]

        if (numberSpikes > 0):
            #print(str(catWeights) + " " + str(neuronNumber))
            #get the weights for each category based on test spikes
            #print("fred" + str(neuronNumber))
            for cat in range (0,10):
                    divisor = neuronCatMatrix[NUMBER_CAT_NEURONS][cat]
                    catWeight = numberSpikes/divisor
                    catWeight = catWeight* neuronCatMatrix[neuronNumber][cat]
                    catWeights[cat] = catWeights[cat] + catWeight
    #print(catWeights)

    #answerCategory = oddResolve(numberZeros,numberOnes,numberTwos)
    #answerCategory = defaultResolve(numberZeros,numberOnes,numberTwos)
    answerCategory = resolve(catWeights)
    
    return (answerCategory)
        
        
def categorise (neuronCatMatrix,allSpikes,labels):
    numberCorrect = 0
    for epoch in range (0,NUMBER_EPOCHS):
        startTime = epoch*EPOCH_LENGTH
        endTime = startTime + EPOCH_LENGTH
        neuronSpikeCounts = getSpikesPerNeuronPerEpoch(allSpikes,startTime,endTime)
        #print (neuronSpikeCounts)
        category = getCategory(neuronCatMatrix,neuronSpikeCounts)
        correctAnswer = labels[epoch]
        if (category == correctAnswer):
                numberCorrect += 1

        print (str(category) + " " + str(correctAnswer) + " " + str(numberCorrect))


    return numberCorrect

#--main
args = sys.argv
numberArgs = args.__len__()
if (numberArgs == 4):
    neuronMatrixFileName = args[1]
    spikeFileName = args[2]
    dataFileName = args[3]
elif (numberArgs == 2):
    fileName = args[1]
elif (numberArgs == 3):
    fileName = args[1]
    spikeFileName = args[2]
else:
    fileName = "spikesPerCat.pkl"
    spikeFileName = "cat.sp"

neuronCatMatrix = getNeuronCatMatrix(neuronMatrixFileName)
neuronCatMatrix = addSums(neuronCatMatrix)
#print (neuronCatMatrix)

allTestSpikes = readSpikes(spikeFileName)
#print (allTestSpikes)
data = readMNISTDataFile(dataFileName)
labels = data[1]
print(labels)


result = categorise (neuronCatMatrix,allTestSpikes,labels)
print ("defaults " + str(defaults))

print (result)
