"""
the iaf_cond_exp_adaptive has variables and parameters.
  Parameters:
  The following parameters can be set in the status dictionary.
V_reset [mV]  Reset Potential
t_ref [ms]  Refractory period
tau_m [ms]  leakage-caused membrane relaxation time constant
C_m [pF]  Membrane Capacitance
E_ex [mV]  Excitatory reversal Potential
E_in [mV]  Inhibitory reversal Potential
E_L [mV]  Leak reversal Potential (aka resting potential)
tau_synE [ms]  Synaptic Time Constant Excitatory Synapse
tau_synI [ms]  Synaptic Time Constant for Inhibitory Synapse
I_e [pA]  constant external input current
tc_theta [ms]  threshold relaxation time constant 10,000,000.0
Theta_plus [mV]  Theta += Theta_plus on each spike 0.05
Theta_rest [mV]  Resting value of the dynamic threshold -72.0


  Dynamic state variables:
r [integer]  counts number of tick during the refractory period
V_m [mV]  membrane potential
V_th [mV]  dynamic firing threshold


in the plastic synapse 
alpha real    Asymmetry parameter (scales depressing increments as alpha*lambda)
nu (lambda) deafult = 0.01 (diehl = 0.01)
tau_plus  ms      Time constant of STDP window, potentiation (tau_minus defined in postsynaptic neuron)
lambda    real    Step size
mu_plus   real    Weight dependence exponent, potentiation
mu_minus  real    Weight dependence exponent, depression
Wmax      real    Maximum allowed weight
"""
import pyNN.nest as sim
from pyNN.nest import native_cell_type
from pyNN.nest import native_synapse_type

import sys
import random
import pickle #for saving synapses

"""From diehl_network.py
        V_th=(-50.0, -40.0),  # (exc_neurons, inh_neurons)
        t_ref=(5.0, 2.0),  # (exc_neurons, inh_neurons)
        tau_m=(100.0, 10.0),  # (exc_neurons, inh_neurons)
"""
TRAIN_FILE_NAME = "undoneinputData/Dat2a.txt"
TEST_FILE_NAME = "undoneinputData/Dat2b.txt"
#NUMBER_EXAMPLES = 2
#NUMBER_EXAMPLES = 10
#NUMBER_EXAMPLES = 281
NUMBER_EXAMPLES = 562
#NUMBER_EXAMPLES = 2810
NUMBER_INPUTS = 64
NUMBER_CATEGORY_NEURONS = 200
NUMBER_INHIBITORY_NEURONS = 200
PERCENTAGE_INPUT_CATEGORY_SYNAPSE = 0.6
PERCENTAGE_INPUT_INHIBITORY_SYNAPSE = 0.05
INITIAL_EXC_FIRING_THRESHOLD = -50.0
INITIAL_INH_FIRING_THRESHOLD = -40.0
SYNAPSE_ALPHA = 0.55
WMAX = 10.0

INPUT_RATE_MULTIPLIER = 50.0
INPUT_INHIB_MEDIAN = 0.0002 #from args
INPUT_INHIB_DELAY_MEDIAN = 5.0

INPUT_CAT_MEDIAN = 10.0
INPUT_CAT_DELAY_MEDIAN = 10.0

#INHIB_CAT_MEDIAN = -0.017
INHIB_CAT_MEDIAN = -0.01
CAT_INHIB_MEDIAN = 0.0104


#---Below here are functions to test the system---
EPOCH_LENGTH = 400
REST_BETWEEN_EPOCHS = 50
RUN_DURATION = EPOCH_LENGTH*NUMBER_EXAMPLES
TIME_STEP = 1.0
#TIME_STEP = 0.1
#set in process args
THETA_PLUS = 0.05

TAU_SYNE = 1.0
EXC_TC_THETA = 10000000.0
EXC_V_RESET = -65.0
E_IN = -100.0
EXC_E_L = -65.0

EXC_C_M = 100.0
EXC_T_REF = 5.0
EXC_TAU_M = 100.0

INH_T_REF = 2.0
INH_TAU_M = 10.0
INH_V_RESET = -45.0
INH_C_M = 10.0
INH_E_L = -60.0
INH_TC_THETA=1e+20
INH_THETA_PLUS = 0.0
INH_THETA_REST = -40.0
INH_CELL_PARAMS = {'t_ref':INH_T_REF,'tau_m':INH_TAU_M,
                   'V_reset':INH_V_RESET,'C_m':INH_C_M,'tau_synE':TAU_SYNE,
                   'E_in':E_IN,"E_L": INH_E_L,'tc_theta': INH_TC_THETA,
                   'Theta_plus':INH_THETA_PLUS,
                   'Theta_rest': INH_THETA_REST
                   }

EXC_CELL_PARAMS = {'t_ref':EXC_T_REF,'tau_m':EXC_TAU_M,'tc_theta': EXC_TC_THETA,
                   'V_reset':EXC_V_RESET,'C_m':EXC_C_M,'tau_synE':TAU_SYNE,
                   'E_in':E_IN,"E_L": EXC_E_L,
                   'Theta_plus':THETA_PLUS}
SEED = 1
#global variables
random_gen = random  #setup the random number generator
random_gen.seed(SEED)

def processArgs(args):
    global TRAIN_FILE_NAME
    global TEST_FILE_NAME
    global INPUT_RATE_MULTIPLIER
    global INPUT_INHIB_MEDIAN 
    #global THETA_PLUS  when you add something from exc_cell_params or inh
    #into the sweep, you need to update it here too.
    #global EXC_CELL_PARAMS 
    
    numberArgs = args.__len__()
    print(str(args) + " " + str(numberArgs))

    if (numberArgs < 2):
        print("Error: call mnistTest with some arguments")
    else:
        runType = args[1]
        synapseFile = args[2]
        TEST_FILE_NAME = args[3]
        rateMultiplier = args[4]
        INPUT_RATE_MULTIPLIER = float(rateMultiplier)
        inputToInhibMedian = args[5]
        INPUT_INHIB_MEDIAN = float(inputToInhibMedian)


    if (runType == "test"):
        return (synapseFile,TEST_FILE_NAME)
    else:  #train
        if (numberArgs == 6):
            TRAIN_FILE_NAME = args[2]
            TEST_FILE_NAME = args[3]
            rateMultiplier = args[4]
            INPUT_RATE_MULTIPLIER = float(rateMultiplier)
            inputToInhibMedian = args[5]
            INPUT_INHIB_MEDIAN = float(inputToInhibMedian)
        else:
            print ("need parameters more parameters for mnistTest")
            fileName = "inputData/irisDat2a.txt"
        
    #EXC_CELL_PARAMS = {'t_ref':EXC_T_REF,'tau_m':EXC_TAU_M,
    #'tc_theta': EXC_TC_THETA,
    #'V_reset':EXC_V_RESET,'C_m':EXC_C_M,'tau_synE':TAU_SYNE,
    #'E_in':E_IN,"E_L": INH_E_L,
    #'Theta_plus':THETA_PLUS}
    return None

    
def processLine(lineString):
    dataLine = lineString.split(',')
    #print (dataLine)
    #print (len(dataLine))
    label = int (dataLine[len(dataLine)-1])
    #print(label)
    features = []
    for featureNumber in range (0,len(dataLine)-1):
        nextFeature = int(dataLine[featureNumber])
        features = features + [nextFeature]
        
    label = [label]

    #print(features)
    #print(label)
    
    return (features,label)

def normalize(feature):
    min = 20.0
    
    for featureSetNumber in range (0,75):
        if (feature[featureSetNumber] < min):
            min = feature[featureSetNumber]

    outputFeature = []
    for featureSetNumber in range (0,75):
        outputFeature = outputFeature + [feature[featureSetNumber]-min]

    max = -45.0
    for featureSetNumber in range (0,75):
        if (outputFeature[featureSetNumber] > max):
            max = outputFeature[featureSetNumber]

    result = []
    for featureSetNumber in range (0,75):
        result = result + [outputFeature[featureSetNumber]/max]

    return (result)
        
        
def normalizeFeatures(features):
    feature1 = []
    feature2 = []
    feature3 = []
    feature4 = []
    for featureSetNumber in range (0,75):
        feature1 = feature1+ [features[featureSetNumber][0]]
        feature2 = feature2+ [features[featureSetNumber][1]]
        feature3 = feature3+ [features[featureSetNumber][2]]
        feature4 = feature4+ [features[featureSetNumber][3]]

    feature1 = normalize(feature1)
    feature2 = normalize(feature2)
    feature3 = normalize(feature3)
    feature4 = normalize(feature4)

    result = []
    for featureSetNumber in range (0,75):
        newRow = [feature1[featureSetNumber],feature2[featureSetNumber],
                  feature3[featureSetNumber],feature4[featureSetNumber]]
        result = result + [newRow]

    return result

def readMyMNISTDataFile(fileName):
    fileHandle = open(fileName,'r')
    features = []
    labels = []
    #line = fileHandle.readline()
    for line  in fileHandle:
        feature,label = processLine(line)
        features = features + [feature]
        labels = labels + label
    fileHandle.close()

    #Normalize the features.
    #features = normalizeFeatures(features)
    #print(features)
    print(len(features))
    print(labels)
    print(len(labels))
    return(features,labels)

#get a set of generators
def getGens(inputNumber,inputFeatures,startOffset):
        duration = EPOCH_LENGTH - REST_BETWEEN_EPOCHS
        start = (inputNumber*EPOCH_LENGTH)+startOffset

        #boost rate
        rateMultiplier = INPUT_RATE_MULTIPLIER
        rates = []
        for featureNumber in range (0,len(inputFeatures)):
            value = inputFeatures[featureNumber]* rateMultiplier
            rates = rates + [value]
                
        
        #rate = inputFeatures
        #print (str(start)+ " " + str(rates))
        genParams = {'start':start,'rate':rates,'duration': duration}
        inputGens = sim.Population(NUMBER_INPUTS,sim.SpikeSourcePoisson,genParams)
        return (inputGens)
        
    
def addReverseFeatureData(inputData):
    result = []
    for featureSetNumber in range (0,75):
        row = inputData[featureSetNumber]
        for inputFeatureNumber in range (0,4):
            newItem = row[inputFeatureNumber]
            newItem = 1-newItem
            row = row + [newItem]
        result = result + [row]

    return (result)

def createInputs():        
        #make the input cells
        inputCells=sim.Population(NUMBER_INPUTS,parrot_neuron_type,{})
        inputCells.record({'spikes'})
        return (inputCells)

        #get the train data
        trainData = readMyMNISTDataFile(trainFileName)
        featureData = trainData[0]
        #featureData = addReverseFeatureData(featureData)
        labels = trainData[1]
        numberExamples = len(labels)
        if (NUMBER_EXAMPLES != numberExamples):
                print ("change NUMBER_EXAMPLES_CONSTANT")

        #make the generators and connect them
        for inputExample in range (0,NUMBER_EXAMPLES):
            inputGens = getGens(inputExample,featureData[inputExample],5)
            sim.Projection(inputGens,inputCells,sim.OneToOneConnector())

        #get generators for the categorisation section.
        start = RUN_DURATION+5
        for inputExample in range (0,NUMBER_EXAMPLES):
            inputGens = getGens(inputExample,featureData[inputExample],start)
            sim.Projection(inputGens,inputCells,sim.OneToOneConnector())

        """undone
        #get generators to test on the train
        numberExamples = len(labels)
        start = (RUN_DURATION*2)+5
        for inputExample in range (0,NUMBER_EXAMPLES):
            inputGens = getGens(inputExample,featureData[inputExample],start)
            sim.Projection(inputGens,inputCells,sim.OneToOneConnector())

        """
        #get generators to test on the the test
        inputData = readMyMNISTDataFile(testFileName)
        featureData = inputData[0]
        #featureData = addReverseFeatureData(featureData)
        labels = inputData[1]
        numberExamples = len(labels)
        start = (RUN_DURATION*2)+5
        #start = (RUN_DURATION*3)+5
        for inputExample in range (0,NUMBER_EXAMPLES):
            inputGens = getGens(inputExample,featureData[inputExample],start)
            sim.Projection(inputGens,inputCells,sim.OneToOneConnector())


#Make the generators for the input, and make synapses to the inputCells
def connectGeneratorsToInput(inputCells,inputData,startTime):
    featureData = inputData[0]
    labels = inputData[1]
    numberExamples = len(labels)
    start = startTime + 5
    for inputExample in range (0,NUMBER_EXAMPLES):
            inputGens = getGens(inputExample,featureData[inputExample],start)
            sim.Projection(inputGens,inputCells,sim.OneToOneConnector())

        
def createNeurons(fTrain,trainFileName,testFileName):
        #make the input cells and their generators
        inputCells = createInputs()

        if (fTrain):
            trainData = readMyMNISTDataFile(trainFileName)
            connectGeneratorsToInput(inputCells,trainData,0)
            
        else:
            inputData = readMyMNISTDataFile(testFileName)
            connectGeneratorsToInput(inputCells,inputData,0)

        #make categorisation cells.
        categorisationCells=sim.Population(NUMBER_CATEGORY_NEURONS,
                                           diehl_neuron_type,EXC_CELL_PARAMS)

        categorisationCells.record({'spikes'})
        #categorisationCells.record({'spikes','V_m'})
        categorisationCells.initialize(V_th=INITIAL_EXC_FIRING_THRESHOLD)

        #inhibitory cells
        inhibitoryCells=sim.Population(NUMBER_INHIBITORY_NEURONS,
                                       diehl_neuron_type,INH_CELL_PARAMS)
        #inhibitoryCells.record({'spikes','V_m'})
        inhibitoryCells.record({'spikes'})
        inhibitoryCells.initialize(V_th=INITIAL_INH_FIRING_THRESHOLD)

        #print("inhibitory parameters")
        
        #print(inhibitoryCells)
        #print(inhibitoryCells[0])
        #print("V_reset " + str(inhibitoryCells[0].V_reset))
        #print("t_ref " + str(inhibitoryCells[0].t_ref))
        #print("tau_m " + str(inhibitoryCells[0].tau_m))
        #print("C_m " + str(inhibitoryCells[0].C_m))
        #print("E_ex " + str(inhibitoryCells[0].E_ex))
        #print("E_in " + str(inhibitoryCells[0].E_in))
        #print("E_L " + str(inhibitoryCells[0].E_L))
        #print("tau_synE " + str(inhibitoryCells[0].tau_synE))
        #print("tau_synI " + str(inhibitoryCells[0].tau_synI))
        #print("I_e " + str(inhibitoryCells[0].I_e))
        #print("Theta_plus " + str(inhibitoryCells[0].Theta_plus))
        #print("tc_theta " + str(inhibitoryCells[0].tc_theta))
        #print("Theta_rest " + str(inhibitoryCells[0].Theta_rest))
        return [inputCells,categorisationCells,inhibitoryCells]

def connectInputAndInhibit(inputCells,inhibCells):
    connector = []
    for fromNumber in range(0,NUMBER_INPUTS):
        for toNumber in range(0,NUMBER_INHIBITORY_NEURONS):
            makeSynapseProb = random_gen.random()
            if (makeSynapseProb < PERCENTAGE_INPUT_INHIBITORY_SYNAPSE):
                weight = random_gen.random()
                weight = weight * INPUT_INHIB_MEDIAN

                delay = random_gen.random()
                delay = (delay*INPUT_INHIB_DELAY_MEDIAN) 
                if (delay < TIME_STEP): delay = TIME_STEP
                connector = connector + [(fromNumber,toNumber,weight,delay)]

        listConnector = sim.FromListConnector(connector)
        synapses = sim.Projection(inputCells, inhibCells, listConnector)

def readSynapses(fileName):
        fileHandle = open(fileName,"rb")
        inputToCatSynapses = pickle.load(fileHandle)
        fileHandle.close()

        #Need to multiply the weight * 1000
        newSynapses = []
        for synapse in inputToCatSynapses:
                newWeight = synapse[2]
                #newWeight = newWeight*100
                newDelay = synapse[3]
                synapse = (synapse[0],synapse[1],newWeight,newDelay)
                newSynapses = newSynapses + [synapse]

        listConnector = sim.FromListConnector(newSynapses)
        return (listConnector)
        

def connectInputCat(inputCells,catCells,fPlastic,synapseFileName):
    print(fPlastic)
    if (fPlastic):
        diehl_synapse_type = native_synapse_type('stdp_nn_symm_synapse')
        weightMedian = INPUT_CAT_MEDIAN
        connector = []
        for fromNumber in range(0,NUMBER_INPUTS):
            for toNumber in range(0,NUMBER_CATEGORY_NEURONS):
               makeSynapseProb = random_gen.random()
               if (makeSynapseProb < PERCENTAGE_INPUT_CATEGORY_SYNAPSE):

                   weight = random_gen.random()
                   weight = weight * weightMedian 
                   
                   delay = random_gen.random()
                   #delay = (delay*INPUT_CAT_DELAY_MEDIAN) + 1.0
                   delay = (delay*INPUT_CAT_DELAY_MEDIAN) 
                   if (delay < TIME_STEP): delay = TIME_STEP
               
                   connector = connector + [(fromNumber,toNumber,weight,delay)]
        listConnector = sim.FromListConnector(connector)
        plasticSynapses = sim.Projection(inputCells, catCells, listConnector,
#                   synapse_type=sim.StaticSynapse())                           
                   synapse_type=diehl_synapse_type(alpha=SYNAPSE_ALPHA,
                        Wmax=WMAX,mu_plus=0.0, mu_minus=0.0,tau_plus=20.0))

        print ("Synapse Params")
        print (plasticSynapses)
        print ("Tau Plus" + str (plasticSynapses.get("tau_plus",format="list")[0]))
        print ("Mu Plus" + str(plasticSynapses.get("mu_plus",format="list")[0]))
        print ("Mu Minus" + str(plasticSynapses.get("mu_minus",format="list")[0]))
        print ("Wmax" + str(plasticSynapses.get("Wmax",format="list")[0]))
        print ("lambda" + str(plasticSynapses.get("lambda",format="list")[0]))
        #dumpSynapses(connector,"initSynapses.pkl")
        #plasticSynapseList = plasticSynapses.get(['weight','delay'],format='list')
        #dumpSynapses(plasticSynapseList,"newSynapses.pkl")



        return plasticSynapses
    else:
        #listConnector=readSynapses("synapses.pkl")
        listConnector=readSynapses(synapseFileName)
        staticSynapses = sim.Projection(inputCells, catCells, listConnector,
                   synapse_type=sim.StaticSynapse())
        return staticSynapses

def connectCategoryAndInhibit(catCells,inhibCells):
    catToInhibConnector = []
    catToInhibWeightMedian = CAT_INHIB_MEDIAN
    """
    """
    for neuronNumber in range(0,NUMBER_CATEGORY_NEURONS):
        catToInhibWeight = random_gen.random()
        catToInhibWeight = catToInhibWeight * catToInhibWeightMedian 
        catToInhibConnector = catToInhibConnector + [
            (neuronNumber,neuronNumber,catToInhibWeight,TIME_STEP)]



    
    inhibToCatConnector = []
    inhibToCatWeightMedian = INHIB_CAT_MEDIAN
    #get the connection lists between the populations
    for catNumber in range(0,NUMBER_CATEGORY_NEURONS):
            for inhibNumber in range(0,NUMBER_INHIBITORY_NEURONS):
                if (catNumber != inhibNumber):
                    inhibToCatWeight = random_gen.random()
                    inhibToCatWeight = inhibToCatWeight * inhibToCatWeightMedian 
                    inhibToCatConnector = inhibToCatConnector + [
                        (inhibNumber,catNumber,inhibToCatWeight,TIME_STEP)]
                
                
    #print (catToInhibConnector)

    #Make the list connectors and the actual synapses
    catToInhibListConnector = sim.FromListConnector(catToInhibConnector)
    sim.Projection(catCells, inhibCells,catToInhibListConnector,
                   synapse_type=sim.StaticSynapse())    

    inhibToCatListConnector = sim.FromListConnector(inhibToCatConnector)
    sim.Projection(inhibCells,catCells,inhibToCatListConnector,
                   synapse_type=sim.StaticSynapse())    

def replacePlasticSynapses(plasticSynapses,plasticSynapseList,inputCells,catCells):
    #undone assure that these 0ed synapses have little of no effect or
    #figure out how to remove them
    #print (plasticSynapses)
    #print (plasticSynapses.size())
    plasticSynapses.set(weight= 0.0)
    
    connector = plasticSynapseList
    listConnector = sim.FromListConnector(connector)
    #print (connector)
    ps = sim.Projection(inputCells, catCells, listConnector)

def dumpSynapses(synapseList, fileName):
    fileHandle = open(fileName,"wb")
    pickle.dump (synapseList,fileHandle)
    fileHandle.close()
    
    
##--main---
fTrain = False
args = sys.argv
runArgs = processArgs(args)
if (runArgs == None):
    fTrain = True
else:
    synapseFileName = runArgs[0]

print(str(INPUT_INHIB_MEDIAN) + " " + TRAIN_FILE_NAME + " " + TEST_FILE_NAME + " " +
      str(INPUT_RATE_MULTIPLIER))
print ("input cat median " + str(INPUT_CAT_MEDIAN))
print ("input inhib median " + str(INPUT_INHIB_MEDIAN))
print ("input rate multiplier " + str(INPUT_RATE_MULTIPLIER))

#link to  cell types that are non-standard nest
parrot_neuron_type = native_cell_type('parrot_neuron')
diehl_neuron_type = native_cell_type('iaf_cond_exp_adaptive')

sim.setup(timestep= TIME_STEP, min_delay=TIME_STEP, max_delay=10.0, debug=0)

#Make neurons and make them recordable
#Align features by spikeSourcePoisson and parrot neurons.  
pops = createNeurons(fTrain,TRAIN_FILE_NAME,TEST_FILE_NAME)
inputCells = pops[0]
categorisationCells= pops[1]
inhibitoryCells= pops[2]

connectInputAndInhibit(inputCells,inhibitoryCells)

#setup input to cat connections
connectCategoryAndInhibit(categorisationCells,inhibitoryCells)


if (fTrain):
    #I think this needs to be last so that it will work for read
    #with the earlier weights being the same.
    plasticSynapses = connectInputCat(inputCells,categorisationCells,True,None)
    print ("start run l")
else:
    staticSynapses = connectInputCat(inputCells,categorisationCells,False,
                                     synapseFileName)
    print ("start run 1 test")
print (RUN_DURATION)
sim.run(RUN_DURATION)

#print results
inputCells.write_data('input.pkl',['spikes'])
categorisationCells.write_data('cat.pkl',['spikes'])
#categorisationCells.write_data('vcat.pkl',['V_m'])
inhibitoryCells.write_data('inhib.pkl',['spikes'])

if (fTrain):
    #train run
    plasticSynapseList = plasticSynapses.get(['weight','delay'],format='list')
    dumpSynapses(plasticSynapseList,"synapses.pkl")
"""

    #replace the plastic synapses with static synapses.
    replacePlasticSynapses(plasticSynapses,plasticSynapseList,inputCells,categorisationCells)

    #get spikes for setting the categories.
    print ("start run c")
    sim.run(RUN_DURATION)
    categorisationCells.write_data('cat2.pkl',['spikes'])
    inputCells.write_data('input.pkl',['spikes'])

    #get spikes for test on the test
    sim.run(RUN_DURATION)
    categorisationCells.write_data('cat4.pkl',['spikes'])
    inputCells.write_data('input.pkl',['spikes'])



    #categorisationCells.write_data('vcat.pkl',['V_m'])

    #get spikes for testing on the train
    print ("start run Test On Train")
    sim.run(RUN_DURATION)
    categorisationCells.write_data('cat3.pkl',['spikes'])
    inputCells.write_data('input.pkl',['spikes'])
"""


