"""
This generates the neuronCats.pkl file, which is the list of the categories
associated with the 100 neurons.  It generates the neuronCats.pkl file.  This
is just a cateogry (or -3) for each neuron.
"""

import sys
import numpy as np
import pickle

EPOCH_LENGTH = 400

def readSpikes (fileName):
    spikes = []
    fileHandle = open(fileName,'r')
    line = fileHandle.readline()
    while not (line == ""):
        arguments = line.split()
        neuronNumber = int(arguments[0])
        time = float(arguments[1])
        spike = (neuronNumber,time)
        spikes = spikes + [spike]
        line = fileHandle.readline()
    fileHandle.close()

    return spikes

def getNeuronsFiring(spikes):
    neuronSpikeCounts = np.zeros(100)
    for spike in spikes:
        neuron = spike[0]
        neuronSpikeCounts[neuron] = neuronSpikeCounts[neuron] + 1

    numberNeuronsFiring = 0
    for neuron in range (0,100):
        if (neuronSpikeCounts[neuron] > 0):
                numberNeuronsFiring += 1
    return (numberNeuronsFiring)
        

def getSpikesPerNeuron(spikes):
    neuronSpikeCounts = np.zeros(100)
    for spike in spikes:
        neuron = spike[0]
        neuronSpikeCounts[neuron] = neuronSpikeCounts[neuron] + 1

    return (neuronSpikeCounts)

def getSpikesPerEpoch(spikes,startEpoch):
    epochSpikeCounts = np.zeros(75)
    for spike in spikes:
        time = spike[1]
        epoch = int(time/EPOCH_LENGTH)
        epoch = epoch-startEpoch
        #print(epoch)
        epochSpikeCounts[epoch] = epochSpikeCounts[epoch] + 1
    return (epochSpikeCounts)

#categories are 0,1,2 and they appear at regular epochs.
def getSpikesPerEpochCat(spikes,cat):
    neuronSpikeCounts = np.zeros(100)
    for spike in spikes:
        neuron = spike[0]
        time = spike[1]
        epoch = int(time/EPOCH_LENGTH)
        testCategory = epoch%3
        if (testCategory == cat):
                neuronSpikeCounts[neuron] = neuronSpikeCounts[neuron] + 1
    return (neuronSpikeCounts)



#return a matrix with the spikes perNeuronXepoch in the cells, the epoch
#sums in the last column, and category sums in the last row.
def getCombinedCategories(cat0,cat1,cat2):
    combinedCats = np.zeros((100+1,4))
    cat0Sum = 0
    cat1Sum = 0
    cat2Sum = 0
    for neuron in range (0,100):
       combinedCats[neuron][0] = cat0[neuron]
       combinedCats[neuron][1] = cat1[neuron]
       combinedCats[neuron][2] = cat2[neuron]
       combinedCats[neuron][3] = cat0[neuron] + cat1[neuron] + cat2[neuron]
       cat0Sum += cat0[neuron]
       cat1Sum += cat1[neuron]
       cat2Sum += cat2[neuron]
       
    combinedCats[100][0] = cat0Sum
    combinedCats[100][1] = cat1Sum
    combinedCats[100][2] = cat2Sum
    combinedCats[100][3] = cat0Sum  + cat1Sum + cat2Sum
    return (combinedCats)
       

def getCategories(combinedCats):
    neuronCats = []
    zeros = 0
    ones = 0
    twos = 0
    for neuron in range (0,100):
            value = -3
            cat0Number = combinedCats[neuron][0]
            cat1Number = combinedCats[neuron][1]
            cat2Number = combinedCats[neuron][2]
            if (cat0Number > cat1Number) and (cat0Number > cat2Number):
                    value = 0
                    zeros += 1
            elif (cat1Number > cat0Number) and (cat1Number > cat2Number):
                    value = 1
                    ones += 1
            elif (cat2Number > cat0Number) and (cat2Number > cat1Number):
                    value = 2
                    twos += 1
            neuronCats = neuronCats + [value]

    #print(str(zeros) + " " + str(ones) +" " + str(twos))
    return (neuronCats)

def dumpSpikesPerCat(spikesPerCat):
    fileHandle = open("spikesPerCat.pkl",'wb')
    pickle.dump(spikesPerCat,fileHandle)
    fileHandle.close()

def dumpNeuronCats(neuronCats):
    fileHandle = open("neuronCats.pkl",'wb')
    pickle.dump(neuronCats,fileHandle)
    fileHandle.close()
    

#--main
args = sys.argv
numberArgs = args.__len__()
if (numberArgs == 2):
    fileName = args[1]
else:
    fileName = "catCat.sp"

spikes = readSpikes(fileName)
#print (spikes)
#print(getSpikesPerNeuron(spikes))
cat0 = getSpikesPerEpochCat(spikes,0)
cat1 = getSpikesPerEpochCat(spikes,1)
cat2 = getSpikesPerEpochCat(spikes,2)
combinedCats = getCombinedCategories(cat0,cat1,cat2)
#print (combinedCats)
dumpSpikesPerCat(combinedCats)

neuronCats = getCategories(combinedCats)
dumpNeuronCats(neuronCats)
print(neuronCats)
    
print(getNeuronsFiring(spikes))
