"""
This generates the neuronMNISTCats.pkl file, which is the spikes per category
for each category neuron.  It generates the neuronMNISTCats.pkl file.  
"""

import sys
import numpy as np
import pickle 

EPOCH_LENGTH = 400
#NUMBER_EPOCHS = 281
NUMBER_EPOCHS = 562
#NUMBER_CAT_NEURONS = 100
NUMBER_CAT_NEURONS = 200

def processLine(lineString):
    dataLine = lineString.split(',')
    #print (dataLine)
    #print (len(dataLine))
    label = int (dataLine[len(dataLine)-1])
    #print(label)
    features = []
    for featureNumber in range (0,len(dataLine)-1):
        nextFeature = int(dataLine[featureNumber])
        features = features + [nextFeature]
        
    label = [label]

    #print(features)
    #print(label)
    
    return (features,label)

#Read in the mnist csv file returning a data matrix and label vector.
def readMNISTDataFile(fileName):
    fileHandle = open(fileName,'r')
    features = []
    labels = []
    for line  in fileHandle:
        feature,label = processLine(line)
        features = features + [feature]
        labels = labels + label
    fileHandle.close()

    #print(labels)
    #print(len(labels))
    return(features,labels)

def readSpikes (fileName):
    spikes = []
    fileHandle = open(fileName,'r')
    line = fileHandle.readline()
    while not (line == ""):
        arguments = line.split()
        neuronNumber = int(arguments[0])
        time = float(arguments[1])
        spike = (neuronNumber,time)
        spikes = spikes + [spike]
        line = fileHandle.readline()
    fileHandle.close()

    return spikes

def getSpikesPerNeuron(spikes):
    for spike in spikes:
        neuron = spike[0]
        neuronSpikeCounts[neuron] = neuronSpikeCounts[neuron] + 1

    return (neuronSpikeCounts)

def getSpikesPerEpoch(spikes,startEpoch):
    epochSpikeCounts = np.zeros(75)
    for spike in spikes:
        time = spike[1]
        epoch = int(time/200)
        epoch = epoch-startEpoch
        #print(epoch)
        epochSpikeCounts[epoch] = epochSpikeCounts[epoch] + 1
    return (epochSpikeCounts)

#Loop through all the spikes and create the neuronEpoch matrix
def getSpikePerNeuronPerEpoch(spikes):
    neuronSpikeCounts = np.zeros((NUMBER_CAT_NEURONS,NUMBER_EPOCHS))
    for spike in spikes:
        neuron = spike[0]
        time = spike[1]
        epoch = int(time/EPOCH_LENGTH)
        neuronSpikeCounts[neuron,epoch] += 1

    return (neuronSpikeCounts)


#undone
def translateEpochsToCats(neuronEpochMatrix,labels):
    neuronCatCounts = np.zeros((NUMBER_CAT_NEURONS,10))
    for epoch in range (0,NUMBER_EPOCHS):
        label = labels[epoch]
        for neuron in range (0,NUMBER_CAT_NEURONS):
            spikesPerEpochPerNeuron = neuronEpochMatrix[neuron][epoch]
            neuronCatCounts[neuron][label] += spikesPerEpochPerNeuron 

    #print(neuronCatCounts)
    return(neuronCatCounts)


#return a matrix with the spikes perNeuronXepoch in the cells, the epoch
#sums in the last column, and category sums in the last row.
def getCombinedCategories(cat0,cat1,cat2):
    combinedCats = np.zeros((100+1,4))
    cat0Sum = 0
    cat1Sum = 0
    cat2Sum = 0
    for neuron in range (0,100):
       combinedCats[neuron][0] = cat0[neuron]
       combinedCats[neuron][1] = cat1[neuron]
       combinedCats[neuron][2] = cat2[neuron]
       combinedCats[neuron][3] = cat0[neuron] + cat1[neuron] + cat2[neuron]
       cat0Sum += cat0[neuron]
       cat1Sum += cat1[neuron]
       cat2Sum += cat2[neuron]
       
    combinedCats[100][0] = cat0Sum
    combinedCats[100][1] = cat1Sum
    combinedCats[100][2] = cat2Sum
    combinedCats[100][3] = cat0Sum  + cat1Sum + cat2Sum
    return (combinedCats)
       

def getCategories(combinedCats):
    neuronCats = []
    zeros = 0
    ones = 0
    twos = 0
    for neuron in range (0,100):
            value = -3
            cat0Number = combinedCats[neuron][0]
            cat1Number = combinedCats[neuron][1]
            cat2Number = combinedCats[neuron][2]
            if (cat0Number > cat1Number) and (cat0Number > cat2Number):
                    value = 0
                    zeros += 1
            elif (cat1Number > cat0Number) and (cat1Number > cat2Number):
                    value = 1
                    ones += 1
            elif (cat2Number > cat0Number) and (cat2Number > cat1Number):
                    value = 2
                    twos += 1
            neuronCats = neuronCats + [value]

    #print(str(zeros) + " " + str(ones) +" " + str(twos))
    return (neuronCats)

def dumpSpikesPerCat(spikesPerCat):
    fileHandle = open("spikesPerMNISTCat.pkl",'wb')
    pickle.dump(spikesPerCat,fileHandle)
    fileHandle.close()

#--main
args = sys.argv
numberArgs = args.__len__()
if (numberArgs == 3):
    dataFileName = args[1]
    fileName = args[2]
elif (numberArgs == 2):
    fileName = args[1]
else:
    fileName = "catCat.sp"

data = readMNISTDataFile(dataFileName)
labels = data[1]

spikes = readSpikes(fileName)
#print (spikes)
neuronEpochMatrix = getSpikePerNeuronPerEpoch(spikes)
print (len(neuronEpochMatrix))
neuronCategoryMatrix = translateEpochsToCats(neuronEpochMatrix,data[1])
combinedCats = neuronCategoryMatrix
print (combinedCats)
dumpSpikesPerCat(combinedCats)

